from typing import Any, Callable, Dict, List, Optional, Tuple

from seaplane.logs import log

"""
Tasks are weird use / mention hybrids: in a deployment
context, they are a map of Seaplane carrier flows and
the connections between them. In an execution context
on the Seaplane platform, they're containers for executable
code that get the results of endpoints or tasks as input and can
emit good things to their downstream flows and endpoints.
"""


class TaskEvent:
    def __init__(self, id: str, input: Any) -> None:
        self.id = id
        self.status = "in_progress"
        self.input = input
        self.output: Optional[Any] = None
        self.error: Optional[Any] = None

    def set_output(self, output: Any) -> None:
        self.output = output
        self.status = "completed"

    def set_error(self, error: Any) -> None:
        self.error = error
        self.status = "error"


class Task:
    def __init__(
        self,
        func: Callable[[Any], Any],
        id: str,
        type: Optional[str] = None,
        model: Optional[str] = None,
        index_name: Optional[str] = None,
        replicas: Optional[int] = 1,
    ) -> None:
        self.func = func
        self.id = id
        self.args: Optional[Tuple[Any, ...]] = None
        self.kwargs: Optional[Dict[str, Any]] = None
        self.type = type
        self.model = model
        self.sources: List[str] = []
        self.name = func.__name__
        self.index_name = index_name
        self.replicas = replicas

    def process(self, *args: Any, **kwargs: Any) -> Any:
        self.args = args
        self.kwargs = kwargs

        log.info(f"Task '{self.id}' processing...")
        return self.func(*self.args, **self.kwargs)

    def called_from(self, sources: List[str]) -> None:
        self.sources = sources

    def print(self) -> None:
        log.info(f"id: {self.id}, replicas: {self.replicas}")



import json
import os
from typing import Any, Dict

import toml

from seaplane.config import config
from seaplane.errors import SeaplaneError
from seaplane.logs import log

from .decorators import context
from .executor import RealTaskExecutor, SchemaExecutor

PROJECT_TOML = "pyproject.toml"


def validate_project() -> None:
    """
    throws an error if the project directory isn't structured in ways
    we expect it to be.
    """
    if not os.path.exists(PROJECT_TOML):
        raise SeaplaneError(f"{PROJECT_TOML} file missing, seaplane init.")

    try:
        project = read_project_file()
        project_name = project["tool"]["poetry"]["name"]
        main = project["tool"]["seaplane"].get("main", None)

        if not os.path.exists(project_name):
            raise SeaplaneError(
                f"source file {project_name} directory missing, \
                    the source code has to live under {project_name} directory."
            )

        if not project_name or not main:
            raise SeaplaneError(f"{PROJECT_TOML} not valid.")
    except Exception as e:
        log.error_exception(e)
        raise SeaplaneError(f"Seaplane project {PROJECT_TOML} configuration is not valid.")


def read_project_file() -> Dict[str, Any]:
    file = open(PROJECT_TOML, "r")
    data = toml.loads(file.read())
    return data


def persist_schema(schema: Dict[str, Any]) -> None:
    if not os.path.exists("build"):
        os.makedirs("build")

    file_path = os.path.join("build", "schema.json")

    with open(file_path, "w") as file:
        json.dump(schema, file, indent=2)


def build() -> Dict[str, Any]:
    log.info("building...")

    validate_project()

    project_config = read_project_file()
    schema: Dict[str, Any] = {"apps": {}}

    context.set_executor(SchemaExecutor())

    for sm in context.apps:
        result = sm.func("entry_point")
        sm.return_source = result

    for sm in context.apps:
        app: Dict[str, Any] = {
            "id": sm.id,
            "entry_point": {"type": sm.type, "parameters": sm.parameters},
            "tasks": [],
            "io": {},
        }

        for c in sm.tasks:
            task = {"id": c.id, "name": c.name, "replicas": c.replicas}

            for source in c.sources:
                if not app["io"].get(source, None):
                    app["io"][source] = [c.id]
                else:
                    app["io"][source].append(c.id)

            app["tasks"].append(task)

        app["io"]["returns"] = sm.return_source
        schema["apps"][sm.id] = app

    schema["carrier_endpoint"] = config.carrier_endpoint
    schema["identity_endpoint"] = config.identify_endpoint

    persist_schema(schema)

    log.info("Apps build successfully!\n")

    context.set_executor(RealTaskExecutor(context.event_handler))

    return {"schema": schema, "config": project_config}



from typing import Any, Optional

import json

from seaplane_framework.flow import processor

from seaplane.logs import log
from seaplane.object import object_store


class TaskContext:
    """TaskContext is what a Task receives."""

    def __init__(self, body: bytes, meta: dict[str, Any]):
        self.body = body
        # For now let's not imply that customers should touch this
        self._meta = meta
        self._object_data: Optional[bytes] = None

    def emit(self, message: bytes, batch_id: int = 1) -> None:
        """Queues message to send once the task completes.
        batch_id: The index with which to refer to this member of the batch."""
        new_meta = self._meta.copy()
        new_meta["_seaplane_batch_hierarchy"] += f".{batch_id}"
        output_msg = processor._Msg(message, new_meta)
        processor.write(output_msg)

    @property
    def object_data(self) -> bytes:
        """
        Attempts to interpret the body of the context
        as an object store event, and to syncronously load
        the associated object and return it.

        May throw exceptions if the message isn't an object store
        notification, if the object store is unavailable, or if
        the object store would otherwise throw an exception (for example,
        if the notification was for a deleted object.)
        """
        if self._object_data is not None:
            return self._object_data

        try:
            msg = json.loads(self.body)
            bucket = msg["Bucket"]
            obj = msg["Object"]
        except (json.JSONDecodeError, KeyError):
            log.error(
                "it doesn't look like this message was from the object store."
                " Make sure your tasks are configured to listen to object store messages"
                " when using context.object_data"
            )
            raise

        self._object_data = object_store.download(bucket, obj)

        return self._object_data



import functools
import json
from typing import Any, Callable, Dict, List, Optional
from warnings import warn

from seaplane.config import config
from seaplane.errors import HTTPError
from seaplane.logs import log

from .app import App, AppEvent
from .event_handler import EventHandler
from .executor import RealTaskExecutor, TaskExecutor
from .task import Task, TaskEvent


def task_to_json(task: Task) -> Dict[str, Any]:
    return {"id": task.id, "replicas": task.replicas}


def app_to_json(app: App) -> Dict[str, Any]:
    return {
        "id": app.id,
        "type": app.type,
        "parameters": app.parameters,
        "tasks": [task_to_json(task) for task in app.tasks],
    }


def apps_json(apps: List[App]) -> Dict[str, Any]:
    return {
        "type": "apps",
        "payload": [app_to_json(app) for app in apps],
    }


class Context:
    """
    Built up from the various @app and @task decorators in the code.
    """

    def __init__(
        self,
        apps: Optional[List[App]] = None,
        tasks: Optional[List[Task]] = None,
    ) -> None:
        self.actual_app_index = -1
        self.event_handler = EventHandler()
        self.task_executor: TaskExecutor = RealTaskExecutor(self.event_handler)

        if apps is None:
            self.apps = []
        else:
            self.apps = apps

        if tasks is None:
            self.tasks = []
        else:
            self.tasks = tasks

    def active_app(self, id: str) -> None:
        for i, app in enumerate(self.apps):
            if app.id == id:
                self.actual_app_index = i
                break

    def add_event(self, event: AppEvent) -> None:
        self.event_handler.add_event(event)

    def update_event(self, event: AppEvent) -> None:
        self.event_handler.update_event(event)

    def task_event(self, task_event: TaskEvent) -> None:
        self.event_handler.task_event(task_event)

    def get_actual_app(self) -> Optional[App]:
        if self.actual_app_index == -1:
            return None

        return self.apps[self.actual_app_index]

    def add_app(self, app: App) -> None:
        if len(self.apps) == 1 and self.apps[0].id == "temporal":
            self.apps[0] = app
        else:
            self.actual_app_index = len(self.apps)
            self.apps.append(app)

        log.debug(f"🧠 App: {app.id}")
        self.event_handler.on_change(apps_json(self.apps))

    def add_task(self, task: Task) -> None:
        log.debug(f"⌛️ Task {task.id} with {task.replicas} replicas")
        self.tasks.append(task)

    def get_task(self, id: str) -> Optional[Task]:
        for c in self.tasks:
            if c.id == id:
                return c

        return None

    def assign_to_active_app(self, task: Task) -> None:
        self.apps[self.actual_app_index].add_task(task)
        app = context.get_actual_app()
        if app is not None:
            log.info(f"⌛️ Assign Task {task.id} to App: {app.id}")
        else:
            log.info(
                f"🔥 Actual App is None, can't assign \
                    Task {task.id} to App"
            )
        self.event_handler.on_change(apps_json(self.apps))

    def set_executor(self, executor: TaskExecutor) -> None:
        self.task_executor = executor


context = Context()


def app(
    type: str = "API",
    parameters: Optional[List[str]] = None,
    id: Optional[str] = None,
    path: Optional[str] = None,
    _func: Optional[Callable[[Any], Any]] = None,
) -> Callable[[Any, Any], Any]:
    if path is not None:
        warn("The app path argument is deprecated", DeprecationWarning, stacklevel=2)
    if parameters is None:
        parameters = []

    def decorator_apps(func: Callable[[Any], Any]) -> Callable[[Any, Any], Any]:
        root_id = id if id is not None else func.__name__.replace("_", "-")
        set_id = config.name_prefix + root_id

        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            context.active_app(set_id)

            args_str = tuple(arg.decode() if isinstance(arg, bytes) else arg for arg in args)
            args_json = json.dumps(args_str)

            event = AppEvent(app_id=set_id, input=args_json)
            context.add_event(event)
            result = None

            try:
                result = func(*args, **kwargs)
                event.set_output(result)
            except HTTPError as err:
                log.error(f"App error: {err}")
                event.set_error(err)

            context.update_event(event)
            return result

        app = App(func=wrapper, id=set_id, type=type, parameters=parameters)
        context.add_app(app)
        return wrapper

    if not _func:
        return decorator_apps  # type: ignore
    else:
        return decorator_apps(_func)


def import_task(_func: Optional[Callable[[Any], Any]], task: Task) -> Callable[[Any, Any], Any]:
    def decorator_task(func: Callable[[Any], Any]) -> Callable[[Any, Any], Any]:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            args_str = tuple(arg.decode() if isinstance(arg, bytes) else arg for arg in args)
            args_json = json.dumps(args_str)

            event = TaskEvent(task.id, args_json)
            context.task_event(event)

            result = task.process(*args, **kwargs)

            event.set_output(result)
            context.task_event(event)

            return func(result)

        return wrapper

    context.add_task(task)

    if not _func:
        return decorator_task  # type: ignore
    else:
        return decorator_task(_func)


def task(
    id: Optional[str] = None,
    type: Optional[str] = None,
    model: Optional[str] = None,
    index_name: Optional[str] = None,
    replicas: Optional[int] = 1,
    _func: Optional[Callable[[Any], Any]] = None,
) -> Callable[[Any, Any], Any]:
    deprecated_args = {"type": type, "model": model, "index_name": index_name}
    for arg in deprecated_args:
        if deprecated_args[arg] is not None:
            warn(f"The task {arg} argument is deprecated", DeprecationWarning, stacklevel=2)

    def decorator_task(func: Callable[[Any], Any]) -> Callable[[Any, Any], Any]:
        root_id = id if id is not None else func.__name__.replace("_", "-")
        task_id = config.name_prefix + root_id

        task = Task(
            func=func,
            type=type,
            model=model,
            id=task_id,
            index_name=index_name,
            replicas=replicas,
        )
        context.add_task(task)

        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            context.assign_to_active_app(task)

            return context.task_executor.execute(task, *args, **kwargs)

        return wrapper

    if not _func:
        return decorator_task  # type: ignore
    else:
        return decorator_task(_func)



"""
command line interface
"""
from argparse import ArgumentParser
from importlib.metadata import version
import json
import os
import re
import requests
from typing import Any, Dict, cast

from cookiecutter.main import cookiecutter

from seaplane.apps.build import read_project_file, validate_project
import seaplane.run_load_dotenv as denv
from seaplane.sdk_internal_utils.http import headers


class CliError(Exception):
    pass


def get_project() -> Dict[str, Any]:
    """
    Reads pyproject.toml
    """
    validate_project()
    return read_project_file()


def cli_deploy(task: str) -> None:
    """
    Deploys an app or task
    """
    project = get_project()
    project_name = project["tool"]["poetry"]["name"]
    main_script = project["tool"]["seaplane"]["main"]
    os.system(f"poetry run python {project_name}/{main_script} deploy {task}")


def cli_destroy() -> None:
    """
    Destroys an app (deletes streams and flows)
    """
    project = get_project()
    project_name = project["tool"]["poetry"]["name"]
    main_script = project["tool"]["seaplane"]["main"]
    os.system(f"poetry run python {project_name}/{main_script} destroy")


def cli_build() -> None:
    """
    Builds an app (writes build/schema.json)
    """
    project = get_project()
    project_name = project["tool"]["poetry"]["name"]
    main_script = project["tool"]["seaplane"]["main"]
    os.system(f"poetry run python {project_name}/{main_script} build")


def cli_status(watch: bool) -> None:
    """
    Provides status of an app's resources.
      If watch is enabled it must be in your path.
    """
    project = get_project()
    project_name = project["tool"]["poetry"]["name"]
    main_script = project["tool"]["seaplane"]["main"]
    if watch:
        os.system(f"watch poetry run python {project_name}/{main_script} status")
    else:
        os.system(f"poetry run python {project_name}/{main_script} status")


def get_token(fd_url: str) -> str:
    """
    Gets a token from Flightdeck
    """
    # Try to get the API key from .env
    api_key = os.environ.get("SEAPLANE_API_KEY", "None")
    if denv.loaded is False or api_key == "None":
        print("Set your API key in .env")
        exit()

    response = requests.post(fd_url, json={}, headers=headers(api_key))
    if response.ok:
        payload = cast(Dict[str, str], response.json())
    else:
        error_body = response.text
        print(f"Bad Access token request code {response.status_code}, error {error_body}")
        exit()
    return payload["token"]


def cli_request(data: str) -> None:
    """
    Performs an endpoint request, writes and prints the request_id
    """
    if data[0] == "@":
        try:
            data_bytes = open(data[1:], "rb").read()
        except Exception:
            print("Cannot load data file.")
            exit()
    else:
        data_bytes = data.encode()

    # Load schema file
    try:
        schema_file = open(os.path.join("build", "schema.json"), "r")
        schema = json.load(schema_file)
    except Exception:
        print("Cannot load build schema. Try moving to the directory where you deploy your app.")
        exit()

    # Get a token
    fd_url = schema["identity_endpoint"] + "/token"
    token = get_token(fd_url)

    # Make the request
    for app in schema["apps"]:
        app_id = schema["apps"][app]["id"]
    carrier_endpoint = schema["carrier_endpoint"]
    url = f"{carrier_endpoint}/endpoints/{app_id}/request"
    bearer = {"Authorization": f"Bearer {token}", "Content-Type": "application/octet-stream"}
    resp = requests.post(url, headers=bearer, data=data_bytes)
    request_id = json.loads(resp.content)["request_id"]

    # save the request_id to a file
    with open(".request_id", "a") as f:
        f.write(f"{request_id}\n")
    print(f"\nrequest_id: {request_id}")


def cli_response(request_id: str) -> None:
    """
    Gets an archive response from the endpoint
    """
    # TODO: Could (with arg?) show a list of the last 5-10 ids to choose
    if request_id == "":
        try:
            with open(".request_id", "r") as f:
                request_id = f.readlines()[-1].strip("\n")
        except Exception:
            print("No request_id provided as argument or in file")

    # Load schema file
    try:
        schema_file = open(os.path.join("build", "schema.json"), "r")
        schema = json.load(schema_file)
    except Exception:
        print("Cannot load build schema. Try moving to the directory where you deploy your app.")
        exit()

    # Get a token
    fd_url = schema["identity_endpoint"] + "/token"
    token = get_token(fd_url)

    # Make the request
    for app in schema["apps"]:
        app_id = schema["apps"][app]["id"]
    carrier_endpoint = schema["carrier_endpoint"]
    url = f"{carrier_endpoint}/endpoints/{app_id}/response/{request_id}/archive"
    bearer = {"Authorization": f"Bearer {token}", "accept": "application/octet-stream"}
    resp = requests.get(url, headers=bearer, params={"pattern": ".>", "format": "json_array"})
    print(f"\nResponse archive for request_id {request_id}:")
    print(resp.content)


def init(project_name: str) -> None:
    """
    Initializes a project using a template
    """
    cookiecutter_template = "https://github.com/seaplane-io/seaplane-app-python-template.git"
    project_directory = "."

    if not project_name.isidentifier():
        suggestion = re.sub("[^A-Za-z0-9_]", "", project_name)
        raise CliError(f'can\'t make a new project named "{project_name}". Try "{suggestion}"')

    extra_context = {"project_slug": project_name, "seaplane_version": version("seaplane")}

    cookiecutter(
        cookiecutter_template,
        output_dir=project_directory,
        no_input=True,  # Disable any interactive prompts
        extra_context=extra_context,
    )

    print(f"🛩️  {project_name} project generated successfully!")


def main() -> None:
    parser = ArgumentParser(prog="seaplane", description="Seaplane Apps command line interface")
    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # Build command
    subparsers.add_parser("build", help="Build command")

    # Deploy command
    deploy_parser = subparsers.add_parser("deploy", help="Deploy command")
    deploy_parser.add_argument(
        "--task",
        type=str,
        help="seaplane deploy which can include a TASK_ID as a parameter to deploy an individual TASK",  # noqa
    )

    # Destroy command
    subparsers.add_parser("destroy", help="Remove your App and associated tasks")

    # Init command
    init_parser = subparsers.add_parser("init", help="Init command")
    init_parser.add_argument("app", help="Seaplane Apps name")

    # Status command
    stat_parser = subparsers.add_parser("status", help="Status of your app and resources")
    stat_parser.add_argument(
        "--watch",
        "-w",
        default=False,
        action="store_true",
        help="Use UNIX watch command to refresh status every 2s",
    )

    # Endpoints commands
    req_parser = subparsers.add_parser("request", help="Send data to the request endpoint")
    req_parser.add_argument(
        "--data", "-d", type=str, help="Request body, @ to load a file, @- for stdin"
    )
    resp_parser = subparsers.add_parser("response", help="Get data from the response endpoint")
    resp_parser.add_argument("--request_id", "-r", type=str, help="Request ID")

    # Version argument
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=version("seaplane")),
    )

    args = parser.parse_args()

    if args.command == "build":
        cli_build()

    elif args.command == "deploy":
        task = ""
        if args.task:
            task = args.task

        cli_deploy(task)

    elif args.command == "destroy":
        cli_destroy()

    elif args.command == "init":
        init(args.app)

    elif args.command == "status":
        cli_status(args.watch)

    elif args.command == "request":
        cli_request(args.data)

    elif args.command == "response":
        request_id = ""
        if args.request_id:
            request_id = args.request_id
        cli_response(request_id)

    else:
        parser.print_help()

    exit(0)


if __name__ == "__main__":
    main()



from importlib.metadata import version
import os
import sys
import traceback

from seaplane_framework.flow import processor

from seaplane.apps.build import build
from seaplane.apps.decorators import context
from seaplane.apps.deploy import deploy, destroy
from seaplane.apps.status import status
from seaplane.apps.task_context import TaskContext
from seaplane.errors import SeaplaneError
from seaplane.logs import log

# importing this because it loads our .env file as a side effect
from seaplane.config import config  # noqa: F401


def _start_task(task_id: str, save_result: bool) -> None:
    task = context.get_task(task_id)

    if not task:
        raise SeaplaneError(
            f"Task {task_id} not found, \
                            make sure the Task ID is correct."
        )

    processor.start()

    while True:
        log.info(f"Task {task.id} waiting for data...")

        # TODO: A read can fail if the incoming data is using
        #       a newer format than this version of `processor`
        #       understands. This would almost certainly be
        #       our fault, but we should still deadletter the message
        #       rather than crashing.
        message = processor.read()

        if "_seaplane_request_id" not in message.meta:
            # This must be the first task in a smartpipe, so we have to get the
            # Endpoints API generated request ID from the incoming nats_subject.
            request_id = message.meta["nats_subject"].split(".")[
                -1
            ]  # The Endpoints API always adds a request ID as the leaf
            message.meta["_seaplane_request_id"] = request_id

            # Similarly let's initialise the batch hierarchy to start out empty
            message.meta["_seaplane_batch_hierarchy"] = ""

        task_context = TaskContext(message.body, message.meta)

        # TODO: Handle task errors gracefully
        try:
            task.process(task_context)
        except Exception as e:
            # At this point, the SDK user's code has thrown an exception
            # there's nothing we can do but log it and move on.
            # NB: We're not returning here, so any existing batch items
            # will still be written.
            error_str = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
            log.error(
                f"Error running Task:\
                \n {error_str}"
            )

        # The task may have written messages (by virtue of TaskContext.emit), so we now
        # flush them down the pipe.
        processor.flush()


def start() -> None:
    """
    *Run Locally*, this function builds, deploys, and manages a customer application.
    Run on seaplane infrastructure with no arguments, this function
    runs a user @task in a loop, pulling information upstream from the seaplane platform.
    """

    log.info(f"\n\n\tSeaplane Apps version {version('seaplane')}\n")

    if len(sys.argv) > 1:
        command = sys.argv[1]

        if command == "build":
            build()
        elif command == "deploy":
            if len(sys.argv) == 3:
                co_id = sys.argv[2]
                deploy(co_id)
            else:
                deploy()
        elif command == "destroy":
            destroy()
        elif command == "status":
            status()
        else:
            log.error(
                f"Found an invalid internal command `{command}`.\n"
                + "Expected one of: \n"
                + " - build\n"
                + " - deploy\n"
                + " - destroy\n"
                + " - status\n"
            )
        return None

    task_id = os.getenv("TASK_ID")

    if not task_id:
        log.debug("ERROR: Could not find TASK_ID value")
        log.error(
            "Executing a Smartpipe workflow from outside a "
            + "Seaplane Deployment is not currently supported!"
        )
        sys.exit(-1)

    log.info(f"Starting Task {task_id} ...")
    save_result = os.getenv("SAVE_RESULT_TASK", "").lower() == "true"
    _start_task(task_id, save_result)
    return None



import hashlib
import json
import os
import shutil
from typing import Any, Dict, List, Optional
from urllib.parse import urlparse
import zipfile

import requests

from seaplane.config import Configuration, config
from seaplane.logs import log
from seaplane.object import ObjectStorageAPI
from seaplane.sdk_internal_utils.http import headers
from seaplane.sdk_internal_utils.token_auth import with_token

from .app import App
from .build import PROJECT_TOML, build
from .decorators import context
from .task import Task

ENDPOINTS_STREAM = "_SEAPLANE_ENDPOINT"

SecretKey = str
SecretValue = str


def create_endpoints_input_subject(app_id: str) -> str:
    return f"{ENDPOINTS_STREAM}.in.{app_id}.*"


def _file_md5(path: str) -> str:
    """
    Gets the MD5 hash of a file by path.
    """

    hasher = hashlib.md5()
    block_size = 4194304  # 4 MB
    with open(path, "rb") as fh:
        while True:
            buffer = fh.read(block_size)
            if not buffer:
                break
            hasher.update(buffer)
    return hasher.hexdigest()


def create_endpoints_output_subject(app_id: str) -> str:
    # The following ${! ... } incantations are Benthos function interpolation
    request_id = '${! meta("_seaplane_request_id") }'
    joined_batch_hierarchy = '${! meta("_seaplane_batch_hierarchy") }'
    return f"{ENDPOINTS_STREAM}.out.{app_id}.{request_id}{joined_batch_hierarchy}"


def create_subject(app_id: str, task_id: str) -> str:
    return f"{app_id}.{task_id}"


def create_carrier_workload_file(
    tenant: str,
    app: App,
    task: Task,
    next_tasks: List[str],
    project_url: str,
    is_first_task: bool,
    has_to_save_output: bool,
) -> Dict[str, Any]:
    if is_first_task:
        if app.type == "stream" and len(app.parameters) >= 1:
            input = app.parameters[0]
        else:  # API
            input = create_endpoints_input_subject(app.id)
    else:
        input = create_subject(app.id, task.id)

    output: Optional[Dict[str, Any]] = None

    if len(next_tasks) > 1:
        output = {
            "broker": {
                "outputs": (
                    {"carrier": {"subject": create_subject(app.id, c_id)}} for c_id in next_tasks
                )
            }
        }
    elif len(next_tasks) == 1:
        output = {
            "carrier": {"subject": create_subject(app.id, next_tasks[0])},
        }
    else:
        if has_to_save_output:
            output = {
                "carrier": {"subject": create_endpoints_output_subject(app.id)},
            }

    max_ack_pending = 2
    if task.replicas:
        max_ack_pending = task.replicas * 2

    workload = {
        "input": {
            "carrier": {
                "subject": input,
                "durable": task.id,
                "queue": task.id,
                "ack_wait": "2m",
                "max_ack_pending": max_ack_pending,
            },
        },
        "processor": {
            "docker": {
                "image": config.runner_image,
                "args": [project_url],
            }
        },
        "output": output,
        "replicas": task.replicas,
    }

    if not os.path.exists(f"build/{task.id}"):
        os.makedirs(f"build/{task.id}")

    with open(f"build/{task.id}/workload.json", "w") as file:
        json.dump(workload, file, indent=2)
        log.debug(f"Created {task.id} workload")

    return workload


@with_token
def create_stream(token: str, name: str) -> None:
    log.debug(f"Creating stream: {name}")
    url = f"{config.carrier_endpoint}/stream/{name}"

    payload: Dict[str, Any] = {"ack_timeout": 20}  # should be long enough for OpenAI
    if config.region is not None:
        payload["allow_locations"] = [f"region/{config.region}"]
    resp = requests.put(
        url,
        json=payload,
        headers=headers(token),
    )
    resp.raise_for_status()


@with_token
def delete_stream(token: str, name: str) -> None:
    log.debug(f"deleting stream: {name}")
    url = f"{config.carrier_endpoint}/stream/{name}"

    resp = requests.delete(
        url,
        headers=headers(token),
    )
    resp.raise_for_status()


def get_secrets(config: Configuration) -> Dict[SecretKey, SecretValue]:
    secrets = {}
    for key, value in config._api_keys.items():
        secrets[key] = value

    return secrets


@with_token
def add_secrets(token: str, name: str, secrets: Dict[SecretKey, SecretValue]) -> None:
    log.debug(f"adding secrets: {name}")
    url = f"{config.carrier_endpoint}/flow/{name}/secrets"

    flow_secrets = {}
    for secret_key, secret_value in secrets.items():
        flow_secrets[secret_key] = {"destination": "all", "value": secret_value}

    resp = requests.put(
        url,
        json=flow_secrets,
        headers=headers(token),
    )
    resp.raise_for_status()


@with_token
def create_flow(token: str, name: str, workload: Dict[str, Any]) -> None:
    log.debug(f"creating flow: {name}")
    url = f"{config.carrier_endpoint}/flow/{name}"
    if config.dc_region is not None:
        url += f"?region={config.dc_region}"

    resp = requests.put(
        url,
        json=workload,
        headers=headers(token),
    )
    resp.raise_for_status()


@with_token
def delete_flow(token: str, name: str) -> None:
    log.debug(f"deleting flow: {name}")

    url = f"{config.carrier_endpoint}/flow/{name}"
    if config.dc_region is not None:
        url += f"?region={config.dc_region}"

    resp = requests.delete(
        url,
        headers=headers(token),
    )
    resp.raise_for_status()


def zip_current_directory(tenant: str, project_name: str) -> str:
    current_directory = os.getcwd()
    zip_filename = f"./build/{tenant}.zip"

    with zipfile.ZipFile(zip_filename, "w", zipfile.ZIP_DEFLATED) as zipf:
        zipf.write(PROJECT_TOML, os.path.relpath(PROJECT_TOML, current_directory))

        env_file = os.environ.get("SEAPLANE_ENV_FILE", ".env")
        if os.path.exists(env_file):
            zipf.write(env_file, os.path.relpath(".env", current_directory))

        for root, _, files in os.walk(f"{current_directory}/{project_name}"):
            for file in files:
                if "__pycache__" in root:
                    continue

                file_path = os.path.join(root, file)
                zipf.write(file_path, os.path.relpath(file_path, current_directory))

    return zip_filename


def upload_project(project: Dict[str, Any], tenant: str) -> str:
    """
    Zips the project directory and pushes it into the Seaplane object store,
    returning a URL that our executor image can use to refer back to the
    project when executing.
    """

    # Step 1: Make sure we have a bucket to dump our project into
    default_bucket_name: str = "seaplane-internal-flows"
    default_bucket_config = {
        "description": "Seaplane bucket used for flow images. Should not be modified directly.",
        "replicas": 3,
        "max_bytes": -1,  # unlimited
        "allow_locations": ["all"],  # TODO: Georestrictions
    }

    obj = ObjectStorageAPI()
    obj._allow_internal = True
    if default_bucket_name not in obj.list_buckets():
        obj.create_bucket(default_bucket_name, default_bucket_config)

    # Step 2: Build the zip file
    project_name: str = project["tool"]["poetry"]["name"]
    project_file = zip_current_directory(tenant, project_name)
    remote_path = project_name + "." + _file_md5(project_file) + ".zip"

    # Step 3: Upload & return
    obj.upload_file(default_bucket_name, remote_path, project_file)

    obj_url = obj.file_url(default_bucket_name, remote_path)
    log.info(f"uploaded project package {obj_url}")

    return obj_url


def print_endpoints(schema: Dict[str, Any]) -> None:
    apps = schema["apps"].keys()
    if len(apps) > 0:
        log.info("\nDeployed Endpoints:\n")
    for app_id in apps:
        entry_point_type = schema["apps"][app_id]["entry_point"]["type"]
        if entry_point_type == "API":
            log.info(
                f"🚀 {app_id} Endpoint: POST https://{urlparse(config.carrier_endpoint).netloc}/v1/endpoints/{app_id}/request"  # noqa
            )
            log.info(
                f"🚀 {app_id} CLI Command: plane endpoints request {app_id} -d <data> OR @<file>"
            )
        entry_point_params = schema["apps"][app_id]["entry_point"]["parameters"]
        if entry_point_type == "stream" and len(entry_point_params) >= 1:
            input = entry_point_params[0]
            log.info(f"🚀 {app_id} using stream subject {input} as entry point")

    if len(apps) > 0:
        print("\n")


def deploy_task(
    tenant: str,
    app: App,
    task: Task,
    schema: Dict[str, Any],
    secrets: Dict[SecretKey, SecretValue],
    project_url: str,
) -> None:
    delete_flow(task.id)

    is_first_task = schema["apps"][app.id]["io"].get("entry_point", None) == [task.id]

    has_to_save_output = schema["apps"][app.id]["io"].get("returns", None) == task.id

    next_tasks = schema["apps"][app.id]["io"].get(task.id, None)

    if next_tasks is None:
        next_tasks = []

    workload = create_carrier_workload_file(
        tenant, app, task, next_tasks, project_url, is_first_task, has_to_save_output
    )

    create_flow(task.id, workload)
    secrets = secrets.copy()
    secrets["TASK_ID"] = task.id
    secrets["SAVE_RESULT_TASK"] = str(has_to_save_output)
    add_secrets(task.id, secrets)

    # Log some useful info about where this is deployed
    #  Note that region info is only included if we have set it
    deploy_info = ""
    if "staging" in config.carrier_endpoint:
        deploy_info += " in staging"
    if config.dc_region is not None:
        deploy_info += f" in {config.dc_region} data center"
    log.info(f"Deploy for task {task.id} done{deploy_info}")


def deploy(task_id: Optional[str] = None) -> None:
    secrets = get_secrets(config)
    if not config._token_api.api_key:
        log.info("API KEY not set. Please set in .env or seaplane.config.set_api_key()")
        return

    shutil.rmtree("build/", ignore_errors=True)

    project = build()
    schema = project["schema"]
    tenant = config._token_api.get_tenant()
    project_url = upload_project(project["config"], tenant)

    if task_id is not None:
        log.info("Deploying task {task_id}")
        for sm in context.apps:
            for c in sm.tasks:
                if c.id == task_id:
                    deploy_task(tenant, sm, c, schema, secrets, project_url)
    else:  # deploy everything
        # Log some useful info about where this is deployed
        #  Note that region info is only included if we have set it
        deploy_info = ""
        if "staging" in config.carrier_endpoint:
            deploy_info += " in staging"
        if config.region is not None:
            deploy_info += f" in {config.region} region"
        log.info(f"Deploying everything{deploy_info}...")
        for sm in context.apps:
            delete_stream(sm.id)
            create_stream(sm.id)

            for c in sm.tasks:
                deploy_task(tenant, sm, c, schema, secrets, project_url)

    print_endpoints(schema)

    log.info("🚀 Deployment complete")


def destroy() -> None:
    if not config._token_api.api_key:
        log.info("API KEY not set. Please set in .env or seaplane.config.set_api_key()")
        return
    build()

    for sm in context.apps:
        delete_stream(sm.id)

        for c in sm.tasks:
            delete_flow(c.id)
